/* som - self organizing map */

/*  Copyright 1999 Miller Puckette.
Permission is granted to use this software for any purpose provided you
keep this copyright notice intact.

THE AUTHOR AND HIS EMPLOYERS MAKE NO WARRANTY, EXPRESS OR IMPLIED,
IN CONNECTION WITH THIS SOFTWARE.
 
This file is downloadable from http://www.crca.ucsd.edu/~msp .
*/

#include "m_pd.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
static t_class *som_class;


#define DIM 11
#define M 5.
#define MINRAD 1
#define MAXRAD 2.1
#define MINRAD2 3.5
#define MAXRAD2 4.5
#define MAXNPTS 1000
#define MAXNMAP 1000
#define DEFNMAP 40

typedef struct _pt
{
    float p_x;
    float p_y;
} t_pt;

typedef struct _som
{
    t_object x_obj;
    t_outlet *x_out1;
    t_outlet *x_out2;
    t_pt x_pt[MAXNPTS];
    int x_n;
    int x_nmap;
    t_pt x_map[MAXNMAP];
} t_som;


int randi(int range)
{
    int n = (int)((double)(random()) *
        (double)(range) / RAND_MAX);
    if (n >= range)
        n = range;
    return (n);
}

static void som_start(t_som *x, t_floatarg nmap)
{
    int i;
    if (nmap < 1)
        nmap = DEFNMAP;
    else if (nmap > MAXNMAP)
        nmap = MAXNMAP;
    x->x_nmap = nmap;
    for (i = 0; i < x->x_nmap; i++)
        x->x_map[i] = x->x_pt[randi(x->x_n)];
}

static void *som_new(void)
{
    int i, j;
    t_som *x = (t_som *)pd_new(som_class);
    x->x_out1 = outlet_new(&x->x_obj, gensym("list"));
    x->x_out2 = outlet_new(&x->x_obj, gensym("list"));
    x->x_n = 0;
    for (i = 0; i < DIM; i++)
        for (j = 0; j < DIM; j++)
    {
        int yes1 = (sqrt((i-M)*(i-M) + (j-M)*(j-M)) > MINRAD &&
                sqrt((i-M)*(i-M) + (j-M)*(j-M)) < MAXRAD);
        int yes2 = (sqrt((i-M)*(i-M) + (j-M)*(j-M)) > MINRAD2 &&
                sqrt((i-M)*(i-M) + (j-M)*(j-M)) < MAXRAD2);
        int yes3 = (i == 5);
        if (yes2 || yes3)
        {
            x->x_pt[x->x_n].p_x = i;
            x->x_pt[x->x_n].p_y = j;
            x->x_n++;
        }
#if 0
        if (!j) fprintf(stderr, "\n");
        fprintf(stderr, (yes? "*": " "));
#endif
    }
    som_start(x, 0);
    return (x);
}

static void som_clear(t_som *x)
{
    x->x_n = 0;
}

static void som_add(t_som *x, t_floatarg f1, t_floatarg f2)
{
    if (x->x_n >= MAXNPTS)
        return;
    x->x_pt[x->x_n].p_x = f1;
    x->x_pt[x->x_n].p_y = f2;
    x->x_n++;
}

static void som_list(t_som *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, bestj;
    float closest;
    t_atom at[3];
    int ntrain = atom_getfloatarg(0, argc, argv);
    float range = atom_getfloatarg(1, argc, argv);
    float weight = atom_getfloatarg(2, argc, argv);
    for (i = 0; i < ntrain; i++)
    {
        int pn = randi(x->x_n), jlo, jhi;
        float px = x->x_pt[pn].p_x, py = x->x_pt[pn].p_y;
        for (j = 0, closest = 1e20, bestj = 0; j < x->x_nmap; j++)
        {
            float dist = (x->x_map[j].p_x - px) * (x->x_map[j].p_x - px)
                + (x->x_map[j].p_y - py) * (x->x_map[j].p_y - py);
            if (dist < closest)
                closest = dist, bestj = j;
        }
        jlo = bestj - range;
        if (jlo < 0)
            jlo = 0;
        jhi = bestj + range + 1;
        if (jhi >= x->x_nmap)
            jhi = x->x_nmap;
        for (j = jlo; j < jhi; j++)
            x->x_map[j].p_x += weight * (px - x->x_map[j].p_x),
            x->x_map[j].p_y += weight * (py - x->x_map[j].p_y);
            
    }
    for (i = 0; i < x->x_n; i++)
    {
        SETFLOAT(at+0, x->x_pt[i].p_x);
        SETFLOAT(at+1, x->x_pt[i].p_y);
        outlet_list(x->x_out2, 0, 2, at);
    }
    for (i = 0; i < x->x_nmap; i++)
    {
        SETFLOAT(at+0, i);
        SETFLOAT(at+1, x->x_map[i].p_x);
        SETFLOAT(at+2, x->x_map[i].p_y);
        outlet_list(x->x_out1, 0, 3, at);
    }
}

void som_setup(void)
{
    som_class = class_new(gensym("som"), (t_newmethod)som_new,
        0, sizeof(t_som), 0, A_DEFFLOAT, 0);
    class_addmethod(som_class, (t_method)som_start, gensym("start"), 
        A_DEFFLOAT, 0);
    class_addmethod(som_class, (t_method)som_clear, gensym("clear"), 0);
    class_addmethod(som_class, (t_method)som_add, gensym("add"), 
        A_FLOAT, A_FLOAT, 0);
    class_addlist(som_class, som_list);
}
